import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.cbook import flatten
from matplotlib.ticker import FuncFormatter, MaxNLocator

from sktime.utils.validation._dependencies import _check_soft_dependencies
from sktime.utils.validation.forecasting import check_interval_df, check_y
from sktime.utils.validation.series import check_consistent_index_type
from sktime.datatypes import convert_to



def plot_series(
    *series,
    labels=None,
    markers=None,
    colors=None,
    title=None,
    xlabel=None,
    ylabel=None,
    ax=None,
	figsize =(15,6),
    pred_interval=None,
	interval_label ='prediction interval'
):
    """Plot one or more time series.

    Parameters
    
    series : pd.Series or iterable of pd.Series
        One or more time series
    labels : list, default = None
        Names of series, will be displayed in figure legend
    markers: list, default = None
        Markers of data points, if None the marker "o" is used by default.
        The length of the list has to match with the number of series.
    colors: list, default = None
        The colors to use for plotting each series. Must contain one color per series
    title: str, default = None
        The text to use as the figure's suptitle
    pred_interval: pd.DataFrame, default = None
        Output of `forecaster.predict_interval()`. Contains columns for lower
        and upper boundaries of confidence interval.

    Returns
    
    fig : plt.Figure
    ax : plt.Axis

    Examples
    
    >>> from sktime.utils.plotting import plot_series
    >>> from sktime.datasets import load_airline
    >>> y = load_airline()
    >>> fig, ax = plot_series(y)  # doctest: +SKIP
    """
    
    _check_soft_dependencies("matplotlib", "seaborn")
    
    for y in series:
        check_y(y)

    series = list(series)
    series = [convert_to(y, "pd.Series", "Series") for y in series]

    n_series = len(series)
    _ax_kwarg_is_none = True if ax is None else False
    # labels
    if labels is not None:
        if n_series != len(labels):
            raise ValueError(
                """There must be one label for each time series,
                but found inconsistent numbers of series and
                labels."""
            )
        legend = True
    else:
        labels = ["" for _ in range(n_series)]
        legend = False

    # markers
    if markers is not None:
        if n_series != len(markers):
            raise ValueError(
                """There must be one marker for each time series,
                but found inconsistent numbers of series and
                markers."""
            )
    else:
        markers = ["o" for _ in range(n_series)]

    # create combined index
    index = series[0].index
    for y in series[1:]:
        # check index types
        check_consistent_index_type(index, y.index)
        index = index.union(y.index)

    # generate integer x-values
    xs = [np.argwhere(index.isin(y.index)).ravel() for y in series]

    # create figure if no Axe provided for plotting
    if _ax_kwarg_is_none:
        fig, ax = plt.subplots(1, figsize=figsize)

    # colors
    if colors is None or not _check_colors(colors, n_series):
        colors = sns.color_palette("colorblind", n_colors=n_series)

    # plot series
    for x, y, color, label, marker in zip(xs, series, colors, labels, markers):
        # scatter if little data is available or index is not complete
        # if len(x) <= 3 or not np.array_equal(np.arange(x[0], x[-1] + 1), x):
        #     plot_func = sns.scatterplot
        # else:
        plot_func = sns.lineplot

        plot_func(x=x, y=y, ax=ax, marker=marker, label=label, color=color)

    # combine data points for all series
    xs_flat = list(flatten(xs))

    # set x label of data point to the matching index
    def format_fn(tick_val, tick_pos):
        if int(tick_val) in xs_flat:
            return index[int(tick_val)]
        else:
            return ""

	
    # dynamically set x label ticks and spacing from index labels
    ax.xaxis.set_major_formatter(FuncFormatter(format_fn))
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
	
	
	# Change the xtickslabels
    reformat_xticks = lambda x: x if x=='' else pd.to_datetime(x).strftime('%Y-%m-%d')
    xticklabels = [reformat_xticks(item.get_text()) for item in ax.get_xticklabels()]
    ax.set_xticklabels(xticklabels)
	
    # Set the figure's title
    if title is not None:
        ax.set_title(title, size="xx-large")

    # Label the x and y axes
    if xlabel is not None:
        ax.set_xlabel(xlabel)

    _ylabel = ylabel if ylabel is not None else series[0].name
    ax.set_ylabel(_ylabel)

    ax.legend(loc='best', frameon=False)
	
    if pred_interval is not None:
        check_interval_df(pred_interval, series[-1].index)
        ax = plot_interval(ax, pred_interval, interval_label)
    
    if _ax_kwarg_is_none:
        return fig, ax
    else:
        return ax

def plot_interval(ax, interval_df, legend_label ='prediction interval'):
    cov = interval_df.columns.levels[1][0]
    var_name = interval_df.columns.levels[0][0]
    ax.fill_between(
        ax.get_lines()[-1].get_xdata(),
        interval_df[var_name][cov]["lower"].astype("float64"),
        interval_df[var_name][cov]["upper"].astype("float64"),
        alpha=0.2,
        color=ax.get_lines()[-1].get_c(),
        label=f"{int(cov * 100)}% {legend_label}",
    )
    ax.legend()
    return ax




# Standard Inports
import pandas as pd
import numpy as np
import os

from typing import Optional, Tuple, List
from collections.abc import Iterable
from tqdm.autonotebook import tqdm
import time

# Plotting packages
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import seaborn as sns
sns.set(style='white', font_scale=1)
plt.rc('axes', titlesize='large')    
plt.rc('axes', labelsize='large')   
plt.rc('xtick', labelsize='large')   
plt.rc('ytick', labelsize='large')   
plt.rc('legend', fontsize='large')   
plt.rc('figure', titlesize='x-large') 
from forecast.utils.plotting import plot_series

# Forcast models
from sktime.forecasting.naive import NaiveForecaster
from sktime.forecasting.statsforecast import (
StatsForecastAutoARIMA,
StatsForecastAutoETS, 
StatsForecastAutoTheta
)
from sktime.forecasting.tbats import TBATS
from sktime.forecasting.fbprophet import Prophet

ForecastingModels = {
"Naive": NaiveForecaster(),
"AutoARIMA": StatsForecastAutoARIMA(),
"AutoETS": StatsForecastAutoETS(),
"AutoTheta": StatsForecastAutoTheta(),
"TBATS": TBATS(),
"Prophet": Prophet(),
}

# Cross Validation
from sktime.forecasting.base import ForecastingHorizon
from sktime.forecasting.model_selection import (
CutoffSplitter, 
ExpandingWindowSplitter, 
temporal_train_test_split
)
from sktime.forecasting.model_evaluation import evaluate
# Performance Metrics
from sktime.performance_metrics.forecasting import (
MeanAbsolutePercentageError, 
MeanSquaredError,
)

# Filter Warnings
from warnings import simplefilter
simplefilter('ignore')

def summary_perf(insample_result_df, 
			 grouper, 
			 y_true_col = 'y_true',
			 y_pred_col = 'y_pred',
			):

	"""
	Compute summary performance metrics for a given forecast.

	Parameters:
		- insample_result_df (DataFrame): A DataFrame containing the forecast results.
		- grouper (str): The column name to group the forecast results for computing summary metrics.
		- y_true_col (str, optional): The column name representing the true values. Default is 'y_true'.
		- y_pred_col (str, optional): The column name representing the predicted values. Default is 'y_pred'.

	Returns:
		DataFrame: A DataFrame containing summary performance metrics, including RMSE and MAPE.
	"""

	rmse = MeanSquaredError(square_root=True)
	mape = MeanAbsolutePercentageError()
	horizon_metrics = pd.concat([insample_result_df.groupby(grouper).apply(lambda x: rmse(x[y_true_col], x[y_pred_col])).rename('RMSE'),
									 insample_result_df.groupby(grouper).apply(lambda x: mape(x[y_true_col], x[y_pred_col])).rename('MAPE')
									], axis=1)
	return horizon_metrics

# Pandas frequencies
pandas_frequency_dict = {
    "D": "daily",
    "B": "business days",
    "W": "weekly (end of week, default on Sunday)",
    "M": "monthly (end of month)",
    "Q": "quarterly (end of quarter)",
    "A": "annual (end of year)",
    "H": "hourly",
    "T": "minutely",
    "S": "secondly",
    "L": "millisecondly",
    "U": "microsecondly",
    "BQ": "business quarterly (business quarter-end)",
    "BA": "business annual (business year-end)",
    "BH": "business hourly (business hour)"
}
##############################################################################
# Master Class Definition and Plot
##############################################################################

class Forecast(object):
    """
    Forecast class for managing and evaluating forecast models.

    Parameters
    
        data : pd.DataFrame
            A DataFrame containing the input data for forecasting.
        depvar_str : str
            The column name representing the dependent variable for forecasting.
        fh : int
            The forecast horizon, i.e., the number of periods ahead to forecast.
        pct_initial_window : float
            The percentage of data used as the initial training window.
        step_length : int
            The step size for expanding window cross-validation.
        forecaster_name : str, optional
            The name of the forecasting model. Default is 'Naive'.
        forecaster : object, optional
            The forecasting model object. Default is None, which will use the model corresponding to forecaster_name.
        exog_l : list, optional
            List of exogenous variables for forecasting. Default is None.
        freq : str, optional
            The frequency of the time series data. Default is 'B' (business days).
        random_state : int, optional
            Random state for reproducibility. Default is 0.

    Attributes
    
        y : pd.Series
            The time series data representing the dependent variable.
        X : pd.DataFrame or None
            The DataFrame containing exogenous variables or None if there are no exogenous variables.
        forecaster_name : str
            The name of the forecasting model used.
        forecaster : object
            The forecasting model object used for forecasting.
        fh : ForecastingHorizon
            The forecast horizon, i.e., the number of periods ahead to forecast.
        initial_window : int
            The size of the initial training window.
        step_length : int
            The step size for expanding window cross-validation.
        cv : ExpandingWindowSplitter
            The cross-validation window used for expanding window validation.
        X_train : pd.DataFrame or None
            The DataFrame containing exogenous variables for the training set or None if there are no exogenous variables.
        X_test : pd.DataFrame or None
            The DataFrame containing exogenous variables for the test set or None if there are no exogenous variables.
        y_train : pd.Series
            The dependent variable values for the training set.
        y_test : pd.Series
            The dependent variable values for the test set.
        is_fitted : bool
            True if the forecaster is fitted, False otherwise.
        is_evaluated : bool
            True if the forecaster is evaluated on the test set, False otherwise.
        rs : numpy.random.RandomState
            Random state for reproducibility.
        plot : ForecastPlot
            An instance of the ForecastPlot class for plotting utility.
    """

    __description = "Time Series Forecast"
    __author = "Amine Raboun - amineraboun@github.io"

    # Initializer
    def __init__(self,
                 data: pd.DataFrame,
                 depvar_str: str,
                 fh: int,
                 pct_initial_window: float,
                 step_length: int,
                 forecaster_name: str = 'Naive',
                 forecaster: Optional[object] = None,
                 exog_l: Optional[list] = None,
                 freq: str = 'B',
                 random_state: int = 0) -> None:
        """
        Initialize the Forecast instance.

        Parameters
        
            data : pd.DataFrame
                A DataFrame containing the input data for forecasting.
            depvar_str : str
                The column name representing the dependent variable for forecasting.
            fh : int
                The forecast horizon, i.e., the number of periods ahead to forecast.
            pct_initial_window : float
                The percentage of data used as the initial training window.
            step_length : int
                The step size for expanding window cross-validation.
            forecaster_name : str, optional
                The name of the forecasting model. Default is 'Naive'.
            forecaster : object, optional
                The forecasting model object. Default is None, which will use the model corresponding to forecaster_name.
            exog_l : list, optional
                List of exogenous variables for forecasting. Default is None.
            freq : str, optional
                The frequency of the time series data. Default is 'B' (business days).
            random_state : int, optional
                Random state for reproducibility. Default is 0.
        """
        
        self.__init_test(data, depvar_str, exog_l, freq)
        self.y, self.X = self.__clean_data(data=data, 
                                       depvar_str=depvar_str,
                                       exog_l=exog_l,
                                       freq=freq)
        # Forecasting Model Parameters
        self.forecaster_name = forecaster_name

        if forecaster is None:
            # Name of the forecaster should be known only if the forecaster is null
            _err_msg = f'unsupported forecaster! Known forcasters are {list(ForecastingModels.keys())}'
            assert forecaster_name in ForecastingModels.keys(), _err_msg
            self.forecaster = ForecastingModels[forecaster_name]
        else:
            self.forecaster = forecaster

        # Forecasting parameters
        #1. horizon
        self.fh = ForecastingHorizon(np.arange(1, fh+1))
        #2. Initial training window
        self.initial_window = int(len(self.y)*pct_initial_window)
        #3. Step size
        self.step_length = step_length

        #4. Declare the cross validation window
        self.cv = ExpandingWindowSplitter(initial_window=self.initial_window,
                                          step_length=self.step_length,
                                          fh=self.fh)

        # Create the train test sets
        if self.X is None:
            self.X_train =  None
            self.X_test = None
            self.y_train, self.y_test = temporal_train_test_split(
                y=self.y,
                train_size=self.initial_window)
        else:
            self.y_train, self.y_test, self.X_train, self.X_test = temporal_train_test_split(
            y=self.y, 
            X=self.X,
            train_size=self.initial_window)

        self.is_fitted = False
        self.is_evaluated = False
        self.rs = np.random.RandomState(random_state) 
        # Plots
        self.plot = self.__plot()


    def split_procedure_summary(self) -> dict:
        """
        Generate a summary of the cross-validation procedure.

        Returns
        
            dict:
                A dictionary containing details of the cross-validation procedure, including the number of folds, initial window size, step length, and forecast period.
        """

        _n_splits = cv.get_n_splits(self.y)
        cutoffs = [_train[0] for (_train, _test) in self.cv.split(self.y.index)]
        _split_proc= {'Number of Folds': _n_splits,
                      'Initial Window Size': cv.initial_window,
                      'Step Length': cv.step_length,
                      'Forecast period': len(cv.fh),
                      'First Cutoff': cutoffs[0],
                      'Last Curoff': cutoffs[-1]
                     }
        return _split_proc    

    def __plot(self): #-> ForecastPlot
        """
        Create an instance of the ForecastPlot class for plotting utility.

        Returns:
            ForecastPlot:
                An instance of the ForecastPlot class.
        """
        return ForecastPlot(self)  

    def fit(self, 
            on: str = 'all', 
            fh: Optional[ForecastingHorizon] = None
           ): #-> ForecastFit
        """
        Fit the forecaster and compute insample results.

        Parameters:
            on : str, optional
                Either 'train' or 'all'. Sample on which the model is fitted. By default, it is fitted on the entire sample.
            fh : ForecastingHorizon, optional
                Forecast horizon.

        Returns:
            ForecastFit:
                An instance of the ForecastFit class containing the fitted model and insample performance metrics.
        """

        if fh is None:
            fh =  self.fh

        if on == 'all':
            self.forecaster.fit(y=self.y, X=self.X, fh=fh)
            self.is_fitted = True

        elif on=='train':
            self.forecaster.fit(y=self.y_train, X=self.X_train, fh=fh)
            self.is_fitted = False

        else: 
            on_values =['all', 'train']
            raise ValueError(f'argument takes 2 possible values {on_values}')

        return ForecastFit(self)  

    def evaluate(self): #-> ForecastEval
        """
        Evaluate the forecaster out-of-sample.

        Returns:
            ForecastEval:
                An instance of the ForecastEval class containing the out-of-sample evaluation results.
        """

        self.eval = ForecastEval(self)  
        self.is_evaluated = True
        return self.eval

    def predict(self, 
                X: Optional[pd.DataFrame] = None, 
                fh: Optional[ForecastingHorizon] = None, 
                coverage: float = 0.9
               ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Generate predictions (Average and confidence interval) using the fitted model.

        Parameters
        
            X : pd.DataFrame, optional
                Exogenous variables for forecasting. Default is None and takes the exogenous variables defined at instantiation.
            fh : ForecastingHorizon, optional
                Forecast horizon. Default is None and takes the horizon defined at instantiation.
            coverage : float, optional
                The coverage of the confidence interval. Default is 0.9.

        Returns
        
            Tuple[pd.DataFrame, pd.DataFrame]:
                A tuple containing the predictions and the confidence intervals.
        """
        if self.is_fitted==False:
            print("\nModel not fitted yet, or fitted on training sample alone")
            print("Fitting the model on the whole sample ...")
            self.fit(on='all', fh=fh)

        if fh is None:
            fh =  self.fh
        if X is None:
            X = self.X

        y_pred = self.forecaster.predict(X=X, fh=fh)
        y_pred_ints = self.forecaster.predict_interval(X=X, fh=fh, coverage=coverage)

        return y_pred, y_pred_ints

    def update(self, 
               newdata: pd.DataFrame, 
               fh: Optional[ForecastingHorizon] = None, 
               coverage: float = 0.9, 
               refit: bool = False
              ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Update cutoff value to forecast new dates.
        Possibility to refit the model.

        Parameters
        
            newdata : pd.DataFrame
                The new data containing the same columns as the original data.
            fh : ForecastingHorizon, optional
                Forecast horizon. Default is None and takes the horizon defined at instantiation.
            coverage : float, optional
                The coverage of the confidence interval. Default is 0.9.
            refit : bool, optional
                If True, the model will be refitted on the new training data. Default is False.

        Returns
        
            Tuple[pd.DataFrame, pd.DataFrame]:
                A tuple containing the updated predictions and the updated confidence intervals.
        """

        new_y, new_X = self.__clean_data(data=newdata, 
                                           depvar_str=self.depvar,
                                           exog_l = self.exog_l,
                                           freq = self.freq)

        self.forecaster.update(y=new_y, X=new_X, update_params=refit)
        y_pred, y_pred_ints = self.predict(X=new_X, fh=fh, coverage=coverage)
        return y_pred, y_pred_ints


    def __init_test(self,
                    data: pd.DataFrame, 
                    depvar_str: str, 
                    exog_l: Optional[list], 
                    freq: str
                   ) -> None:
        """
        Perform checks for the provided data, dependent variable, exogenous variables, and frequency.

        Parameters
        
            data : pd.DataFrame
                A DataFrame containing the input data for forecasting.
            depvar_str : str
                The column name representing the dependent variable for forecasting.
            exog_l : list or None
                List of exogenous variables for forecasting.
            freq : str
                The frequency of the time series data.

        Returns
        
            None

        Raises
        
            AssertionError:
                If the provided data is not a DataFrame, or if depvar_str is not a valid column in the data,
                or if exog_l is not None or an iterable.
            AssertionError:
                If any column in exog_l is not present in the data.
            AssertionError:
                If freq is not a recognized pandas frequency.
        """

        assert isinstance(data, pd.DataFrame), 'the data should be a dataframe'

        assert depvar_str in data.columns, 'the dependent varible must be a column in the data'
        self.depvar = depvar_str

        cond1 = exog_l is None
        cond2 = isinstance(exog_l, Iterable)
        assert cond1 or cond2, 'exog_l is either None, meaning no X or an iterable object'
        if cond2:
            assert all([c in data.columns for c in exog_l]), 'not all columns are not in the data'
            self.exog_l = exog_l
        else:
            self.exog_l = exog_l

        assert freq in pandas_frequency_dict.keys(), f'Not a pandas recognized frequency. List of pandas frequencies:\n{pandas_frequency_dict}'
        self.freq = freq

        return None

    def __clean_data(self, 
                     data: pd.DataFrame, 
                     depvar_str: str, 
                     exog_l: Optional[list],
                     freq: str
                    ) -> Tuple[pd.Series, pd.DataFrame]:
        """
        Reformat the data taking into account the requested frequency.

        Returns the X and y for the rest of forecasting operations.
        """

        # Declare and stage the variable to forecast
        y = data[depvar_str].dropna().resample(freq).last().ffill()

        # List of Exogenous variables if any
        if exog_l is None:
            X = None
        else:
            assert all([c in data.columns for c in exog_l]), 'not all columns are not in the data'
            X = data[exog_l].resample(freq).last().ffill()
            X = X.loc[y.index]

        return y, X


class ForecastPlot:
    """
    Plotting utility class for Forecast.

    Parameters
    
        LF : Forecast
            An instance of the Forecast class.

    Returns:
        None
    """

    def __init__(self, LF: Forecast):
        self.__dict__.update(LF.__dict__)


    def plot_train_test(self,
                        labels: List[str] = ["y_train", "y_test"],
                        xlabel: Optional[str] = None,
                        ylabel: Optional[str] = None,
                        title: str = 'Train-Test sets',
                        ax: Optional[plt.Axes] = None,
                        figsize: Tuple[float, float] = (15, 6)) -> None:
        """
        Plot the dependent variable separating the train from the test windows.

        Parameters
        
            labels : List[str], optional
                Labels for the plot. Default is ["y_train", "y_test"].
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Train-Test sets'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).

        Returns
        
            None
        """
        return plot_series(self.y_train, self.y_test, 
                           labels =labels, 
                           ax = ax, 
                           xlabel = xlabel,
                           ylabel = ylabel,
                           title = title,
                           figsize = figsize)


    def plot_cv_procedure(self,
                          ax: Optional[plt.Axes] = None,
                          labels: List[str] = ["Window", "Forecasting horizon"],
                          ylabel: str = "Window number",
                          xlabel: str = "Time",
                          title: str = "Cross Validation Procedure") -> None:
        """
        Plot the cross-validation procedure.

        Parameters
        
            ax : plt.Axes, optional
                The Axes object for the plot.
            labels : List[str], optional
                Labels for the plot. Default is ["Window", "Forecasting horizon"].
            ylabel : str, optional
                Label for the y-axis. Default is "Window number".
            xlabel : str, optional
                Label for the x-axis. Default is "Time".
            title : str, optional
                The title of the plot. Default is "Cross Validation Procedure".

        Returns
        
            None
        """

        _train_windows, _test_windows = self._get_windows()
        self._plot_windows(_train_windows, _test_windows,
                           ax = ax,
                           labels = labels,
                           xlabel = xlabel, 
                           ylabel = ylabel,
                           title = title)


    def plot_prediction(self,
                        y_pred: pd.Series,
                        y_pred_ints: Optional[pd.DataFrame] = None,
                        interval_label: str = 'prediction interval',
                        labels: List[str] = ["y_train", "y_pred"],
                        xlabel: Optional[str] = None,
                        ylabel: Optional[str] = None,
                        title: str = 'Prediction',
                        ax: Optional[plt.Axes] = None,
                        figsize: Tuple[float, float] = (15, 6)):
        """
        Plot the forecast predictions and the confidence intervals.

        Parameters
        
            y_pred : pd.Series
                The predicted values.
            y_pred_ints : pd.DataFrame, optional
                The DataFrame containing the prediction intervals.
            interval_label : str, optional
                Label for the prediction interval. Default is 'prediction interval'.
            labels : List[str], optional
                Labels for the plot. Default is ["y_train", "y_pred"].
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Prediction'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).

        Returns
        
            None
        """

        y = self.y
        y_train = y.loc[y.index<y_pred.index[0]]
        zoom_y_train = y_train.iloc[-5*len(y_pred):]	
        return plot_series(zoom_y_train, y_pred,
                           labels= labels,
                           pred_interval=y_pred_ints,	
                           interval_label ='prediction interval',
                           xlabel = xlabel,
                           ylabel = ylabel,
                           title = title, 
                           ax = ax,
                           figsize = figsize
                          )

    def plot_prediction_true(self,
                             y_pred: pd.Series,
                             y_pred_ints: Optional[pd.DataFrame] = None,
                             interval_label: str = 'prediction interval',
                             labels: List[str] = ["y_train", "y_true", "y_pred"],
                             xlabel: Optional[str] = None,
                             ylabel: Optional[str] = None,
                             title: str = 'Prediction',
                             ax: Optional[plt.Axes] = None,
                             figsize: Tuple[float, float] = (15, 6)):
        """
        Plot the forecast predictions, true values, and the confidence intervals.

        Parameters
        
            y_pred : pd.Series
                The predicted values.
            y_pred_ints : pd.DataFrame, optional
                The DataFrame containing the prediction intervals.
            interval_label : str, optional
                Label for the prediction interval. Default is 'prediction interval'.
            labels : List[str], optional
                Labels for the plot. Default is ["y_train", "y_true", "y_pred"].
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Prediction'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).

        Returns
        
            None
        """

        y = self.y
        y_train = y.loc[y.index<y_pred.index[0]]
        zoom_y_train = y_train.iloc[-5*len(y_pred):]
        true_pred_idx = np.intersect1d(y.index, y_pred.index)
        err_msg = 'No overlap between true values and predicted values.\nIf you want to plot prediction alone use the function plot_prediction'
        assert len(true_pred_idx)>0, err_msg
        y_true = self.y[true_pred_idx]

        return plot_series(zoom_y_train, y_true, y_pred,
                           pred_interval=y_pred_ints,
                           interval_label ='prediction interval',
                           labels=labels, 
                           xlabel = xlabel,
                           ylabel = ylabel,
                           title = title, 
                           ax = ax,
                           figsize = figsize)

    def _plot_windows(self,
                      train_windows: List[np.ndarray],
                      test_windows: List[np.ndarray],
                      ax: plt.Axes,
                      labels: List[str],
                      ylabel: str,
                      xlabel: str,
                      title: str) -> None:
        """
        Visualize training and test windows.

        Parameters
        
            train_windows : List[np.ndarray]
                List of training window indices.
            test_windows : List[np.ndarray]
                List of test window indices.
            ax : plt.Axes
                The Axes object for the plot.
            labels : List[str]
                Labels for the plot.
            ylabel : str
                Label for the y-axis.
            xlabel : str
                Label for the x-axis.
            title : str
                The title of the plot.

        Returns
        
            None
        """
        assert len(labels)==2, 'wrong number of labels'
        simplefilter("ignore", category=UserWarning)

        def get_y(length, split):
            # Create a constant vector based on the split for y-axis."""
            return np.ones(length) * split
        reformat_xticks = lambda x: x if x=='' else pd.to_datetime(x).strftime('%Y-%m-%d')

        n_splits = len(train_windows)
        n_timepoints = len(self.y)
        len_test = len(test_windows[0])

        train_color, test_color = sns.color_palette("colorblind")[:2]

        if ax is None:
            f, ax = plt.subplots(figsize=plt.figaspect(0.3))

        for i in range(n_splits):
            train = train_windows[i]
            test = test_windows[i]

            ax.plot(
                np.arange(n_timepoints), get_y(n_timepoints, i), marker="o", c="lightgray"
            )
            ax.plot(
                train,
                get_y(len(train), i),
                marker="o",
                c=train_color,
                label=labels[0],
            )
            ax.plot(
                test,
                get_y(len_test, i),
                marker="o",
                c=test_color,
                label=labels[1],
            )
        #import pdb; pdb.set_trace()
        xticklabels = [int(item) for item in ax.get_xticks()[1:-1]]
        xticklabels = [''] + [self.y.index[i].strftime('%Y-%m-%d') for i in xticklabels] + ['']
        ax.invert_yaxis()
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        ax.set(
            title=title,
            ylabel=ylabel,
            xlabel=xlabel,
            xticklabels = xticklabels,
        )
        # remove duplicate labels/handles
        handles, labels = [(leg[:2]) for leg in ax.get_legend_handles_labels()]
        ax.legend(handles, labels, frameon=False)

        if ax is None:			
            return f, ax
        else:
            return ax


    def _get_windows(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        Generate windows for the cross-validation procedure.

        Returns
        
            tuple:
                A tuple containing two lists: train_windows and test_windows.
        """
        _train_windows = []
        _test_windows = []
        _y_index = self.y.index 
        for i, (_train, _test) in enumerate(self.cv.split(_y_index)):
            _train_windows.append(_train)
            _test_windows.append(_test)
        return _train_windows, _test_windows 


##############################################################################
# Model Fit and Insample Performance
##############################################################################
class ForecastFit:
    """
    Class for fitting the forecast model and computing insample performance metrics.

    Parameters
    
        Forecast : Forecast
            An instance of the Forecast class.

    Returns
    
        None
    """

    def __init__(self, Forecast: Forecast):
        """
        Initialize the ForecastFit object.

        Parameters
        
            Forecast : Forecast
                An instance of the Forecast class.
        """

        self.__dict__.update(Forecast.__dict__)

        self.plot = self.__plot()


    def insample_predictions(self, nsample: int = 100) -> pd.DataFrame:
        """
        Compute the insample predictions for the fitted model.

        Parameters
        
            nsample : int, optional
                The number of samples to compute. Default is 100.

        Returns
        
            pd.DataFrame
                A DataFrame containing the insample predictions.
        """

        assert self.forecaster.is_fitted, 'Fit the forecast on the training window first before you can evaluate the insample performance'

        _cutoffs = np.random.choice(len(self.y_train)-len(self.fh), size=nsample, replace=False)

        cv_in = CutoffSplitter(cutoffs=_cutoffs, window_length=1, fh=self.fh)

        insample_result = []                
        for i, (intrain, intest) in tqdm(enumerate(cv_in.split_series(self.y_train))):
            fh = ForecastingHorizon(intest.index, is_relative=False)

            # Apply forecaster to predict the past
            in_pred = self.forecaster.predict(fh=fh, X=self.X).rename('y_pred').reset_index()
            in_pred['y_true'] = intest.values
            in_pred['Abs_diff'] = (in_pred['y_true'] - in_pred['y_pred']).abs()
            in_pred['horizon'] = np.arange(1, len(in_pred)+1)
            in_pred['cutoff'] = intrain.index[-1]

            insample_result.append(in_pred)

        insample_result_df = pd.concat(insample_result)
        self.insample_result_df = insample_result_df
        return insample_result_df

    def insample_perf(self) -> dict:
        """
        Compute insample performance metrics (RMSE and MAPE) for the fitted model.

        Returns
        
            dict
                A dictionary containing the computed insample performance metrics.
        """

        if 'insample_result_df' in self.__dict__.keys():
            pass
        else:           
            self.insample_result_df = self.insample_predictions()

        insample_perf_summary = {
            'cutoff': summary_perf(self.insample_result_df, grouper = 'cutoff', y_true_col = 'y_true', y_pred_col = 'y_pred'),
            'horizon': summary_perf(self.insample_result_df, grouper = 'horizon', y_true_col = 'y_true', y_pred_col = 'y_pred')
        }
        self.insample_perf_summary = insample_perf_summary
        return insample_perf_summary

    def __plot(self):
        return ForecastFitPlot(self)

class ForecastFitPlot:
    """
    Plotting utility class for ForecastFit.

    Parameters
    
        LFF : ForecastFit
            An instance of the ForecastFit class.

    Returns
    
        None
    """

    def __init__(self, LFF: ForecastFit):
        """
        Initialize the ForecastFitPlot object.

        Parameters
        
            LFF : ForecastFit
                An instance of the ForecastFit class.
        """

        self.__dict__.update(LFF.__dict__)
        self.LFF = LFF

    def plot_insample_performance(self,
                                  metric: str = 'RMSE',
                                  title: str = 'Insample Performance'):
        """
        Plot the insample performance metrics.

        Parameters
        
            metric : str, optional
                The performance metric to plot. Default is 'RMSE'.
            title : str, optional
                The title of the plot. Default is 'Insample Performance'.
        """

        assert metric in ['RMSE', 'MAPE'], f'{metric} not in summary performance'
        if 'insample_perf_summary' not in self.__dict__.keys():
            insample_perf_summary = self.LFF.insample_perf()

        f, axes = plt.subplots(1,2,figsize=(15,5))
        for i, (_grouper, _df) in enumerate(insample_perf_summary.items()):
            _df[metric].plot(ax= axes[i], style = '-o', title = f'{metric} By {_grouper}')
            axes[i].set_xlabel('')

        plt.suptitle(title)
        plt.tight_layout()                   


##############################################################################
# Model Out of Sample Evaluation
##############################################################################

class ForecastEval:
    """
    Class for evaluating the forecaster out-of-sample.

    Parameters
    
        Forecast : Forecast
            An instance of the Forecast class.

    Returns
    
        None
    """

    def __init__(self, Forecast: Forecast):
        """
        Initialize the ForecastEval object.

        Parameters
        
            Forecast : Forecast
                An instance of the Forecast class.
        """

        self.__dict__.update(Forecast.__dict__)     

        self.scoring_metrics = [MeanSquaredError(square_root=True),
                                MeanAbsolutePercentageError()]
        _rename_metrics = {
            'test_MeanSquaredError':'RMSE', 
            'test_MeanAbsolutePercentageError':'MAPE'
        }
        print(f"\nStart forecaster {self.forecaster_name} evalution....")
        print(" Depending on the forecaster this step may take couple of minutes. Please don't kill the kernel")
        st = time.time()
        self.oos_eval = evaluate(forecaster=self.forecaster, 
                            y=self.y,
                            X = self.X,
                            cv=self.cv,
                            strategy="refit",
                            return_data=True,
                            scoring = self.scoring_metrics,
                            backend ='loky',
                           )
        self.oos_eval = self.oos_eval.set_index('cutoff').sort_index()
        self.oos_eval = self.oos_eval.rename(columns = _rename_metrics)
        et = time.time()
        elapsed_time = et - st
        print(f"Evaluation time: {np.around(elapsed_time / 60,3)} minutes")

        convert_horizon = self.oos_eval.apply(self.__eval_horizon, axis=1)
        self.oos_horizon_df = pd.concat(convert_horizon.values)

        self.plot = self.__plot()


    def summary_results(self) -> pd.DataFrame:
        """
        Generate a summary of out-of-sample forecast results.

        Returns
        
            pd.DataFrame
                A DataFrame containing various summary statistics of the out-of-sample forecasts.
        """

        _summary = {'Number of Folds':  self.oos_eval.shape[0], 
         'Avg Fit time (s)': self.oos_eval.fit_time.mean(),
         'Avg_pred_time (s)': self.oos_eval.pred_time.mean(),
         'Smallest training window': self.oos_eval.len_train_window.min(),
         'Largest training window':self.oos_eval.len_train_window.max(),
         'First cutoff': self.oos_eval.index[0],
         'Last cutoff': self.oos_eval.index[-1],
         'Avg MAPE': self.oos_eval.MAPE.mean(),
         'Avg RMSE': self.oos_eval.RMSE.mean()
        }
        return pd.Series(_summary).to_frame().T

    def summary_horizon(self) -> pd.DataFrame:
        """
        Generate a summary of out-of-sample forecast results per horizon.

        Returns
        
            pd.DataFrame
                A DataFrame containing summary performance metrics (RMSE and MAPE) for each horizon.
        """

        self.oos_horizon_perf = summary_perf(self.oos_horizon_df, grouper='horizon', y_true_col = 'y_test', y_pred_col = 'y_pred')        
        return self.oos_horizon_perf

    def __eval_horizon(self, x):
            _fct = pd.concat([x['y_test'], x['y_pred']], keys=['y_test', 'y_pred'], axis=1)
            _fct['Abs_diff'] = (_fct['y_test'] - _fct['y_pred']).abs()        
            _fct['horizon'] = np.arange(1, len(_fct)+1)    
            _fct['cutoff'] = x.name
            return _fct

    def __plot(self):
        return ForecastEvalPlot(self)

class ForecastEvalPlot:
    """
    Plotting utility class for ForecastEval.

    Parameters
    
        LFE : ForecastEval
            An instance of the ForecastEval class.

    Returns
    
        None
    """

    def __init__(self, LFE: ForecastEval):
        """
        Initialize the ForecastEvalPlot object.

        Parameters
        
            LFE : ForecastEval
                An instance of the ForecastEval class.
        """

        self.__dict__.update(LFE.__dict__)
        self.LFE = LFE

    def plot_oos_score(self,
                       score: str = 'RMSE',
                       xlabel: str = None,
                       ylabel: str = None,
                       title: str = 'Out of Sample Performance - Average on All Horizons',
                       ax: Optional[plt.Axes] = None,
                       figsize: Tuple[float, float] = (15, 6)):
        """
        Plot out-of-sample performance metric historically.

        Parameters
        
            score : str, optional
                The performance metric to plot. Default is 'RMSE'.
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Out of Sample Performance - Average on All Horizons'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).
        """

        assert score in self.oos_eval.columns, 'score not computed'
        if ax is None:
            f, ax = plt.subplots(1,1,figsize=figsize)

        ylabel = score if ylabel is None else ylabel
        xlabel = '' if xlabel is None else xlabel
        self.oos_eval[score].plot(ax = ax, style = '-o')
        ax.set(xlabel = xlabel, ylabel=ylabel, title =title)

        if ax is None:
            return f, ax
        else:
            return ax

    def plot_oos_horizon(self,
                         score: str = 'RMSE',
                         xlabel: str = None,
                         ylabel: str = None,
                         title: str = 'Out of Sample Performance - Average per Horizons',
                         ax: Optional[plt.Axes] = None,
                         figsize: Tuple[float, float] = (15, 6)):
        """
        Plot out-of-sample performance metric per horizon.

        Parameters
        
            score : str, optional
                The performance metric to plot. Default is 'RMSE'.
            xlabel : str, optional
                Label for the x-axis.
            ylabel : str, optional
                Label for the y-axis.
            title : str, optional
                The title of the plot. Default is 'Out of Sample Performance - Average per Horizons'.
            ax : plt.Axes, optional
                The Axes object for the plot.
            figsize : Tuple[float, float], optional
                The figure size. Default is (15, 6).
        """
        if 'oos_horizon_perf' not in self.__dict__.keys():
            oos_horizon_perf = self.LFE.summary_horizon()

        assert score in oos_horizon_perf.columns, 'score not computed'

        if ax is None:
            f, ax = plt.subplots(1,1,figsize=figsize)

        ylabel = score if ylabel is None else ylabel
        xlabel = '' if xlabel is None else xlabel
        
        oos_horizon_perf[score].plot(ax = ax, style = '-o')
        ax.set(xlabel = xlabel, ylabel=ylabel, title =title)

        if ax is None:
            return f, ax
        else:
            return ax




import pandas as pd
import numpy as np
import os
import matplotlib.pyplot as plt
import seaborn as sns

from typing import Union, Optional, Tuple, List, Dict, Any

from warnings import simplefilter
simplefilter('ignore')

from forecast_ar.forecast import *
from forecast_ar.utils.plotting import plot_series

##############################################################################
# Model Selection
##############################################################################

class ForecastModelSelect:

    """
    Class for model selection and comparison based on out-of-sample performance.
    Evaluate how to best combine the different models based on their oos performance.

    The class can be initialized in 2 ways. 
        Method 1: Requires a list of Forecast objects. It is the preferable instantiation if the models have already been evaluated out of sample.

        Method 2: Requires all the arguments taken by Forecast plus a dictionary where the keys are names of forecasts and the values are the forecast_models.

    Parameters
    
        LF_list : list[Forecast], optional
            List of Forecast objects. Default is None.
        models_d : dict, optional
            A dictionary containing various forecasting models for comparison. Default is None.
        mode : str, optional
            The aggregation mode. Default is 'nbest_average_horizon'.
        score : str, optional
            The performance score. Default is 'RMSE'.
        nbest : int, optional
            Number of best models to aggregate. Default is 2.
        kwargs
            Additional keyword arguments to be passed to the Forecast initialization.

    Raises
    
        AssertionError
            If the mode, score, or nbest values are not recognized or do not meet the requirements.
            If both LF_list and models_d are None.
    """

    def __init__(self, 
                 LF_list: Optional[List[Forecast]] = None,
                 models_d: Optional[Dict] = None,
                 mode: str = 'nbest_average_horizon',
                 score: str = 'RMSE',
                 nbest: int = 2,
                 **kwargs: Any
                ) -> None:
        """
        Initialize the ForecastModelSelect instance.

        Parameters
        
            LF_list : list[Forecast], optional
                List of Forecast objects. Default is None.
            models_d : dict, optional
                A dictionary containing various forecasting models for comparison. Default is None.
            mode : str, optional
                The aggregation mode. Default is 'nbest_average_horizon'.
            score : str, optional
                The performance score. Default is 'RMSE'.
            nbest : int, optional
                Number of best models to aggregate. Default is 2.
            **kwargs
                Additional keyword arguments to be passed to the Forecast initialization.

        Raises
        
            AssertionError
                If the mode, score, or nbest values are not recognized or do not meet the requirements.
                If both LF_list and models_d are None.
        """
        self.__init_test(mode, score, nbest)

        if LF_list is not None:
            self.LF_list = LF_list

        else:

            assert models_d is not None, 'You must have either the models_d or the LF_list not empty'
            lf_l = []
            for forecaster_name, forecaster in models_d.items():            
                _lf =  Forecast(
                    forecaster_name = forecaster_name,
                    forecaster = forecaster,
                    **kwargs
                )
                lf_l.append(_lf)
            self.LF_list = lf_l		
        return None

    def add_forecast(self, 
                     lf: Forecast) -> None:
        """
        Add a Forecast Object to the list of models to evaluate.

        Parameters:
            lf : Forecast
                The Forecast object to be added.
        """
        self.LF_list.append(lf)
        return self.LF_list


    def fit(self, 
            on: str = 'all', 
            fh: Optional[ForecastingHorizon] = None, 
            force: bool = False
           ) -> None:
        """
        Fit the forecasting models for all the underlying Forecast objects.

        Parameters
        
            on : str, optional
                The period for the in-sample fitting. Default is 'all'.
            fh : ForecastingHorizon, optional
                The forecasting horizon. Default is None and it takes the value entered at initialization..
            force : bool, optional
                If True, force fitting even if the models are already fitted. Default is False.

        Returns
        
            None
        """
        for _lf in self.LF_list:
            if (_lf.is_fitted) and force ==False:
                pass
            else:
                _lf.fit(on=on, fh=fh)
        return None

    def evaluate(self,
                 force: bool = False
                ) -> Tuple:
        """
        Evaluate the underlying Forecast models out of sample.

        Parameters
        
            force : bool, optional
                If True, force the evaluation even if the models are already evaluated. Default is False.

        Returns
        
            tuple
                A tuple containing the summary of out-of-sample performance per horizon and per model.
        """
        #Step 1: Loop on all models and evaluate them
        _model_evals = []
        for _lf in self.LF_list:
            if (_lf.is_evaluated) and force ==False:
                _model_evals.append(_lf.eval)
                pass
            else:
                _model_evals.append(_lf.evaluate())
        self.eval_models = _model_evals

        # Step 2: Get the OOS Summary of Performance 
        #per horizon and per model
        _sum_h = {}; _sum_T = {}
        for _lf_eval in self.eval_models:            
            _sum_T[_lf_eval.forecaster_name] = _lf_eval.summary_results().squeeze()
            _sum_h[_lf_eval.forecaster_name] = _lf_eval.summary_horizon()

        self.summary_horizon = pd.concat(_sum_h.values(),
                                         axis=1,
                                         keys = _sum_h.keys())
        self.summary_results = pd.concat(_sum_T.values(), 
                                         axis=1,
                                         keys = _sum_T.keys())

        return self.summary_horizon, self.summary_results

    def select_best(self,

                    score: Optional[str] = None,

                    nbest: Optional[int] = None

                   ) -> pd.DataFrame:
        """
        Select the best model based on horizon and overall performance.

        Parameters
        
            score : str, optional
                The performance metric to use for model comparison. Should be either 'RMSE' or 'MAPE'. Default is None and it takes the value entered at initialization.
            nbest : int, optional
                The number of best models to select based on horizon performance. Default is None and it takes the value entered at initialization.

        Returns
        
            pd.DataFrame
                model_rank_perhorizon. Rank of models per horizon based on performance.
        """

        if score is None:
            score = self.score

        if nbest is None:
            nbest = self.nbest

        nevals = len(self.LF_list)
        if nbest > nevals:
            print(f'\nnbest ={nbest} is higher than the number of models evaluated {nevals}')
            print(f'The average performance of best models will be computed on {nevals} models instead')
            nbest = max(nbest, nevals)

        cond1 = 'eval_models' not in self.__dict__.keys()
        cond2 = 'summary_horizon' not in self.__dict__.keys()
        if cond1 & cond2:	
            print('\nTo select best. We have to evaluate the models out of sample')
            print('Run evaluate ...')
            self.evaluate()

        self.best_x_overall = {}
        self.model_rank_perhorizon = {}
        self.avg_oos_horizon = {}
        for score in ['RMSE', 'MAPE']:

            r = self.summary_horizon.unstack().unstack(1)[score].unstack()

            # Select best x models based on their average performance on all horizons 
            best_x_overall = list(r.mean(axis=1).sort_values(ascending=True).index[:nbest])

            # Compute the rank of the models per horizon
            model_rank_perhorizon = r.apply(lambda x: x.nsmallest(nevals).index.tolist(), axis=0)
            model_rank_perhorizon = model_rank_perhorizon.reset_index(drop=True)\
            .rename(index={i: f'Best_{i+1}' for i in range(nevals)})

            # Summary Model average
            avg_oos_horizon = pd.concat([r.mean(), 
                                           r.loc[best_x_overall].mean(),
                                           r.apply(lambda x: x.nsmallest(nbest).mean(), axis=0)
                                          ], 
                                         keys = ['Model Avg (all models)',  
                                          f'Best {nbest} Model (over all)',
                                          f'Best {nbest} Model (per horizon)'], 
                                         axis = 1)

            self.avg_oos_horizon[score] = pd.concat([r.T, avg_oos_horizon], axis=1)
            self.model_rank_perhorizon[score] = model_rank_perhorizon
            self.best_x_overall[score] = best_x_overall

        return self.model_rank_perhorizon[score]


    def oos_per_cutoff(self,
                       score: Optional[str] = None,
                       nbest: Optional[int] = None
                      ) -> pd.DataFrame:
        """
        Calculate and return the out-of-sample performance based on cutoffs.

        Parameters
        
            score : str, optional
                The performance metric to use for model comparison. Should be either 'RMSE' or 'MAPE'. Default is None and it takes the value entered at initialization.
            nbest : int, optional
                The number of best models to select based on horizon performance. Default is None and it takes the value entered at initialization.

        Returns
        
            pd.DataFrame
                A DataFrame containing the out-of-sample performance per horizon and per model based on the specified score and nbest values.
        """
        if score is None:
            score = self.score

        if nbest is None:
            nbest = self.nbest

        nevals = len(self.LF_list)
        if nbest > nevals:
            print(f'\nnbest ={nbest} is higher than the number of models evaluated {nevals}')
            print(f'The average performance of best models will be computed on {nevals} models instead')
            nbest = max(nbest, nevals)

        cond = 'eval_models' not in self.__dict__.keys()
        if cond:
            print('\nTo select best. We have to evaluate the models out of sample')
            print('Run evaluate ...')
            self.evaluate()

        avg_oos_hist = {}
        for score in ['RMSE', 'MAPE']:
            _oos = {_lf_eval.forecaster_name:_lf_eval.oos_eval[score] for _lf_eval in self.eval_models}
            _oos = pd.concat(_oos.values(),  axis=1, keys= _oos.keys())

            nbest_mods = list(_oos.mean().nsmallest(nbest).index)
            _aggs = pd.concat([_oos.mean(axis=1), 
                                             _oos[nbest_mods].mean(axis=1)
                                            ], axis = 1,  keys = ['Model Avg (all models)',  
                                                                  f'Best {nbest} Model (over all)']
                                           )
            avg_oos_hist[score] = pd.concat([_oos, _aggs], axis=1)
        self.avg_oos_hist = avg_oos_hist
        return avg_oos_hist[score]

    def predict(self,
                X: Optional[pd.DataFrame] = None, 
                fh: Optional[ForecastingHorizon] = None,
                coverage: float = 0.9,
                mode: Optional[str] = None,
                score: Optional[str] = None,
                nbest: Optional[int] = None,
                ret_underlying: bool = False
               ) -> Tuple[pd.Series, pd.DataFrame]:
        """
        Make forecasts using the specified aggregation mode.

        Parameters
        
            X : pd.DataFrame, optional
                The exogenous variables used for prediction. Default is None and it takes the value entered at initialization..
            fh : ForecastingHorizon, optional
                The forecasting horizon. Default is None and it takes the value entered at initialization..
            coverage : float, optional
                The prediction interval coverage. Default is 0.9.
            mode : str, optional
                The aggregation mode for predictions. Default is None and it takes the value entered at initialization.. 
                Available values:
                * 'best': The prediction is based on the best model.
                * 'best_horizon': The prediction is based on the best model for each horizon.
                * 'average': The average of the prediction of all models.
                * 'inverse_score': The weighted average prediction, where weights are inversely proportional to the model performance score.
                * 'nbest_average': Average of the n best models. The n is given by the parameter nbest.
                * 'nbest_average_horizon': Average of the n best models for each horizon. The n is given by the parameter nbest.
            score : str, optional
                The performance metric to use for model comparison. Should be either 'RMSE' or 'MAPE'. Default is None and it takes the value entered at initialization..
            nbest : int, optional
                The number of best models to select based on horizon performance. Default is None and it takes the value entered at initialization..
            ret_underlying : bool, optional
                If True, return the underlying predictions and prediction intervals for each model. Default is False.

        Returns
        
            tuple
                A tuple containing the aggregated prediction and prediction intervals.
        """
        preds = {}; pred_ints = {};

        for _lf in self.LF_list:
            _y_pred, _y_pred_ints = _lf.predict(X=X, fh=fh, coverage=coverage)
            preds[_lf.forecaster_name] = _y_pred;
            pred_ints[_lf.forecaster_name] =_y_pred_ints; 

        preds = pd.concat(preds.values(), keys = preds.keys(), axis=1)
        pred_ints = pd.concat(pred_ints.values(), keys = pred_ints.keys(), axis=1)


        return self.__aggregate_pred(mode=mode, preds=preds, pred_ints=pred_ints, score=score, nbest=nbest, ret_underlying=ret_underlying)

    def update(self,
               newdata: pd.DataFrame,
               refit: bool = False, 
               fh: Optional[ForecastingHorizon] = None, 
               coverage: float = 0.9, 
               mode: Optional[str] = None,
               score: Optional[str] = None,
               nbest: Optional[int] = None,
               ret_underlying: bool = False
              ) -> Tuple[pd.Series, pd.DataFrame]:
        """
        Function to update the prediction for all the models and aggregate them based on the specific mode.

        Parameters
        
            newdata : pd.DataFrame
                The new data for updating the predictions.
            refit : bool, optional
                If False, generate the prediction without fitting the model. If True, fit the underlying models again, evaluate them out of sample, select the best models, and aggregate the predictions. Default is False.
            fh : ForecastingHorizon, optional
                The forecasting horizon. Default is None and it takes the value entered at initialization..
            coverage : float, optional
                The prediction interval coverage. Default is 0.9.
            mode : str, optional
                The aggregation mode for predictions. Default is None and it takes the value entered at initialization..
                Available values:
                * 'best': The prediction is based on the best model.
                * 'best_horizon': The prediction is based on the best model for each horizon.
                * 'average': The average of the prediction of all models.
                * 'inverse_score': The weighted average prediction, where weights are inversely proportional to the model performance score.
                * 'nbest_average': Average of the n best models. The n is given by the parameter nbest.
                * 'nbest_average_horizon': Average of the n best models for each horizon. The n is given by the parameter nbest.
            score : str, optional
                The performance metric to use for model comparison. Should be either 'RMSE' or 'MAPE'. Default is None and it takes the value entered at initialization..
            nbest : int, optional
                The number of best models to select based on horizon performance. Default is None and it takes the value entered at initialization. and it takes the value entered at initialization..
            ret_underlying : bool, optional
                If True, return the underlying predictions and prediction intervals for each model. Default is False.

        Returns
        
            tuple
                A tuple containing the aggregated prediction and prediction intervals.
        """
        preds = {}; pred_ints = {};
        for _lf in self.LF_list:
            _y_pred, _y_pred_ints = _lf.update(newdata = newdata, fh=fh, coverage=coverage, refit=refit)
            preds[_lf.forecaster_name] = _y_pred;
            pred_ints[_lf.forecaster_name] =_y_pred_ints; 

        preds = pd.concat(preds.values(), keys = preds.keys(), axis=1)
        pred_ints = pd.concat(pred_ints.values(), keys = pred_ints.keys(), axis=1)

        if refit == False:
            return self.__aggregate_pred(mode=mode, preds=preds, pred_ints=pred_ints, score=score, nbest=nbest, ret_underlying=ret_underlying)
        else:
            self.evalutate(force=True)
            self.select_best(score = score, nbest = nbest)
            return self.__aggregate_pred(mode=mode, preds=preds, pred_ints=pred_ints, score=score, nbest=nbest, ret_underlying=ret_underlying)


    def plot_model_compare(self, 
                           score: str = 'RMSE', 
                           view: str = 'horizon', 
                           xlabel: Optional[str] = None, 
                           ylabel: Optional[str] = None,
                           title: str = 'Out of Sample Performance', 
                           ax: Optional[plt.Axes] = None, 
                           figsize: Tuple[int, int] = (15, 6)
                          ) -> Union[plt.Figure, plt.Axes]:
        """
        Plot a comparison of models based on their out-of-sample performance.

        Parameters
        
            score : str, optional
                The performance metric for comparison. Should be either 'RMSE' or 'MAPE'. Default is 'RMSE'.
            view : str, optional
                The view mode for comparison, either 'horizon' or 'cutoff'. Default is 'horizon'.
            xlabel : str, optional
                The label for the x-axis. Default is an empty string.
            ylabel : str, optional
                The label for the y-axis. Default is an empty string.
            title : str, optional
                The title of the plot. Default is 'Out of Sample Performance'.
            ax : matplotlib.axes._subplots.AxesSubplot, optional
                The matplotlib axes to use for plotting. Default is None.
            figsize : tuple, optional
                The size of the figure in inches (width, height). Default is (15, 6).

        Returns
        
            matplotlib.figure.Figure or matplotlib.axes._subplots.AxesSubplot
                The figure and axes of the plot.
        """
        if view =='horizon':
            if 'avg_oos_horizon' not in self.__dict__.keys():
                self.select_best()
            toplot = self.avg_oos_horizon[score]

        elif view=='cutoff':
            if 'avg_oos_hist' not in self.__dict__.keys():
                self.oos_per_cutoff()
            toplot = self.avg_oos_hist[score]

        else:
            raise ValueError('view can take only 2 values: horizon or cutoff')

        if ax is None:
            f, ax = plt.subplots(1,1,figsize=figsize)
        if ylabel is None:
            ylabel = score
        if xlabel is None:
            xlabel = view

        toplot.plot(ax=ax, style = '-o')
        ax.set(xlabel = xlabel, ylabel=ylabel, title =title)
        ax.legend(frameon=False)

        if ax is None:
            return f, ax
        else:
            return ax


    def plot_prediction(self, 
                        y_pred: pd.Series, 
                        models_preds: Optional[pd.DataFrame] = None, 
                        y_pred_interval: Optional[Tuple] = None, 
                        interval_label: str = 'prediction interval', 
                        aggregation_label: str = 'Model Agg', 
                        xlabel: str = '', 
                        ylabel: str = '', 
                        title: str = 'Prediction', 
                        ax: Optional[plt.Axes] = None, 
                        figsize: Tuple[int, int] = (15, 6)

                       ) -> Union[plt.Figure, plt.Axes]:
        """
        Plot forecast predictions and aggregated prediction.

        Parameters
        
            y_pred : pd.Series
                The aggregated prediction.
            models_preds : pd.DataFrame, optional
                The DataFrame containing predictions from different models. Default is None.
            y_pred_interval : tuple, optional
                Prediction interval for the aggregated prediction. Default is None.
            interval_label : str, optional
                Label for the prediction interval. Default is 'prediction interval'.
            aggregation_label : str, optional
                Label for the aggregated prediction. Default is 'Model Agg'.
            xlabel : str, optional
                Label for the x-axis. Default is an empty string.
            ylabel : str, optional
                Label for the y-axis. Default is an empty string.
            title : str, optional
                Title of the plot. Default is 'Prediction'.
            ax : matplotlib.axes._subplots.AxesSubplot, optional
                Matplotlib axes to use for plotting. Default is None.
            figsize : tuple, optional
                Size of the figure in inches (width, height). Default is (15, 6).

        Returns
        
            matplotlib.figure.Figure or matplotlib.axes._subplots.AxesSubplot
                The figure and axes of the plot.
        """

        y = self.LF_list[0].y
        y_train = y.loc[y.index<y_pred.index[0]]
        zoom_y_train = y_train.iloc[-5*len(y_pred):]
    
        if models_preds is not None:
            model_names = models_preds.columns
            models_preds = [models_preds[c] for c in model_names]
            labels = ['y'] + list(model_names) + [aggregation_label]

            return plot_series(zoom_y_train, *models_preds, y_pred,
                               labels =labels,
                               xlabel =xlabel,
                               ylabel =ylabel,
                               title =title,
                               ax=ax, 
                               figsize =figsize,
                               pred_interval = y_pred_interval,
                               interval_label ='prediction interval')
        else:

            return plot_series(zoom_y_train, y_pred,
                               labels =['y', aggregation_label],
                               xlabel =xlabel,
                               ylabel =ylabel,
                               title =title,
                               ax=ax, 
                               figsize =figsize,
                               pred_interval = y_pred_interval,
                              interval_label ='prediction interval')

    def __aggregate_pred(self, mode, preds, pred_ints, score=None, nbest=None, ret_underlying=False):

        if score is None:
            score = self.score
        if nbest is None:
            nbest = self.nbest
        if mode is None:
            mode = self.mode

        if  (mode !='average') & ('best_x_overall' not in self.__dict__.keys()):
            print('\nIdentify best models and their oos performance ...')
            self.select_best(score = score, nbest = nbest)

        if mode =='best':

            "returns the prediction of the best model"
            y_pred = preds[self.best_x_overall[score][0]]
            y_pred_int = pred_ints[self.best_x_overall[score][0]]


        elif mode =='best_horizon':

            "returns the prediction of the best model for each forecast horizon"
            best_horizon = self.model_rank_perhorizon[score].loc['Best_1'].to_dict()
            y_pred = pd.Series([preds[v].iloc[k-1] for k, v in best_horizon.items()], 
                   index=preds.index)
            y_pred_int = pd.concat([pred_ints[v].iloc[k-1].to_frame().T for k, v in best_horizon.items()])


        elif mode == 'average':

            "returns the average prediction of all models"
            y_pred = preds.mean(axis=1)
            y_pred_int = pred_ints.unstack().unstack(0).mean(axis=1).unstack().T

        elif mode == 'inverse_score':

            "returns the Average models proportionals to performance score"
            _score = self.summary_results.loc[f'Avg {score}'].sort_values()
            _score = _score/_score.sum()

            _weigh_preds =  preds.mul(_score, axis=1).sum(axis=1)
            y_pred = _weigh_preds.astype('float')

            _weigh_preds_ints = 0
            for m, w in _score.to_dict().items():
                _weigh_preds_ints = _weigh_preds_ints+ pred_ints[m]*w
            y_pred_int = _weigh_preds_ints.astype('float')

        elif mode == 'nbest_average':

            "return the average prediction of nbest models"
            y_pred = preds[self.best_x_overall[score]].mean(axis=1)
            y_pred_int = pred_ints[self.best_x_overall[score]]\
            .unstack().unstack(0).mean(axis=1).unstack().T


        elif mode == 'nbest_average_horizon':

            "return the average prediction on the nbest models per horizon"
            best_horizon = self.model_rank_perhorizon[score].iloc[:nbest].T
            best_horizon = best_horizon.apply(lambda x: x.values, axis=1).to_dict()

            y_pred = pd.Series([preds[v].iloc[k-1].mean() for k, v in best_horizon.items()], 
                               index=preds.index)

            y_pred_int = pd.concat([pred_ints[v].iloc[k-1].unstack(0).mean(axis=1).to_frame().T for k, v in best_horizon.items()])
            y_pred_int.index = y_pred.index

        else:
            recog_modes = ['best', 'best_horizon', 'average', 'inverse_score', 
                           'nbest_average', 'nbest_average_horizon']
            _error_msg = f'Aggregation mode not recognized. Recognized prediction aggregation are {recog_modes}'
            raise ValueError(_error_msg)

        if ret_underlying:
            return y_pred, y_pred_int, preds, pred_ints
        else:
            return y_pred, y_pred_int


    def __init_test(self, 
                    mode: str, 
                    score: str,
                    nbest: int
                   ) -> None:
        """
        Check the validity of mode, score, and nbest values.

        Parameters
        
            mode : str
                The aggregation mode.
            score : str
                The performance score.
            nbest : int
                Number of best models to aggregate.

        Raises
        
            AssertionError
                If the mode, score, or nbest values are not recognized or do not meet the requirements.
        """
        assert isinstance(mode, str), 'mode must be a sting'
        recog_modes = ['best', 'best_horizon', 'average', 'inverse_score', 
                           'nbest_average', 'nbest_average_horizon']
        assert mode in recog_modes, f'mode function not implemented!. Recognized modes are {recog_modes}'
        self.mode = mode

        recog_scores = ['RMSE', 'MAPE']
        assert score in ['RMSE', 'MAPE'], f'performance score not implemented!. Recognized scores are {recog_scores}'		
        self.score = score

        assert isinstance(nbest, int), 'n best must be an integer'		
        self.nbest = nbest	
        return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['forecast_ar', 'forecast_ar.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.7.2,<4.0.0',
 'prophet>=1.1.4,<2.0.0',
 'seaborn>=0.12.2,<0.13.0',
 'sktime>=0.21.0,<0.22.0',
 'sphinx>=7.1.2,<8.0.0',
 'statsforecast>=1.5.0,<2.0.0',
 'statsmodels>=0.14.0,<0.15.0',
 'tbats>=1.1.3,<2.0.0',
 'tqdm>=4.66.0,<5.0.0']

setup_kwargs = {
    'name': 'forecast-ar',
    'version': '0.0.0',
    'description': 'Automation of forecast models testing, combining and predicting',
    'long_description': '# forecast\n\n[![License](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)\n\n## Overview\n\nforecast is a Python library built upon the foundation of the sktime library, designed to simplify and streamline the process of forecasting and prediction model aggregation. It provides tools for aggregating predictions from multiple models, evaluating their performance, and visualizing the results. Whether you\'re working on time series forecasting, data analysis, or any other predictive modeling task, forecast offers a convenient and efficient way to handle aggregation and comparison.\n\n## Key Features\n\n- **Model Aggregation:** Easily aggregate predictions from multiple models using various aggregation modes such as best model overall, best model per horizon, inverse score weighted average model, and more.\n- **Out-of-Sample Evaluation:** Evaluate model performance using out-of-sample data and choose the best models based on user-defined performance metrics.\n- **Visualization:** Visualize model performance, aggregated predictions, and prediction intervals with built-in plotting functions.\n- **Flexibility:** Accommodate various aggregation strategies, forecast horizons, and performance metrics to cater to your specific use case.\n\n## Installation\n\nInstall Your Package Name using pip:\n\n```bash\npip install forecast\n```\n\n## Usage\n\n```python\n# Import the necessary classes from your-package-name\ndata = pd.Series(np.cumsum(np.random.normal(0, 1, size=1000)), \n                 index=pd.date_range(end=\'31/12/2022\', periods=1000)).rename(\'y\').to_frame()\n\nfrom forecast.model_select import ForecastModelSelect\nForecastingModels = {\n"Naive": NaiveForecaster(),\n"AutoARIMA": StatsForecastAutoARIMA(),\n"AutoETS": StatsForecastAutoETS(),\n"AutoTheta": StatsForecastAutoTheta(),\n"TBATS": TBATS(),\n"Prophet": Prophet(),\n}\nmodel = ForecastModelSelect(\n            data= data,\n            depvar_str = \'y\',                 \n            exog_l=None,\n            fh = 10,\n            pct_initial_window=0.75,\n            step_length = 25,\n            models_d = ForecastingModels,\n            freq = \'B\',\n            mode = \'nbest_average_horizon\',\n            score = \'RMSE\', \n            nbest = 2)\n\n# compare models\nmodel.select_best(score = \'MAPE\')\n# Visualize model comparison\nmodel.plot_model_compare(score=\'MAPE\', view=\'horizon\')\nmodel.plot_model_compare(score=\'MAPE\', view=\'cutoff\')\n\n# Generate prediction\ny_pred, y_pred_ints, preds, pred_ints =  model.predict(score=\'RMSE\', ret_underlying=True)\n\n# Visualize prediction\nLFMS.plot_prediction(y_pred = y_pred,\n                     models_preds = preds,\n                     y_pred_interval = y_pred_ints, \n                     title = \'Prediction\')\n```\n\n## Documentation\n\nFor detailed information about available classes, methods, and parameters, please refer to the [Documentation](https://amineraboun.github.io/forecast/).\n\n## License\n\nThis project is licensed under the [MIT License](LICENSE).\n\n## Contributing\n\nWe welcome contributions from the community! If you have suggestions, bug reports, or feature requests, please open an issue or submit a pull request. \n\n## Contact\n\nFor queries, support, or general inquiries, please feel free to reach me at [amineraboun@gmail.com](mailto:amineraboun@gmail.com).\n',
    'author': 'amineraboun',
    'author_email': 'amineraboun@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/amineraboun/forecast',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
